@@ -37,6 +37,7 @@
 
 #ifdef HAVE_SYSTEMD
 #  include <systemd/sd-bus.h>
+#  include <limits.h>
 
 #  define SYSTEMD_PROPERTY_PREFIX "org.systemd.property."
 
@@ -106,75 +107,6 @@ property_missing_p (char **missing_properties, const char *property)
   return false;
 }
 
-int
-cpuset_string_to_bitmask (const char *str, char **out, size_t *out_size, libcrun_error_t *err)
-{
-  cleanup_free char *mask = NULL;
-  size_t mask_size = 0;
-  const char *p = str;
-  char *endptr;
-
-  while (*p)
-    {
-      long long start_range, end_range;
-
-      if (*p < '0' || *p > '9')
-        goto invalid_input;
-
-      start_range = strtoll (p, &endptr, 10);
-      if (start_range < 0)
-        goto invalid_input;
-
-      p = endptr;
-
-      if (*p != '-')
-        end_range = start_range;
-      else
-        {
-          p++;
-
-          if (*p < '0' || *p > '9')
-            goto invalid_input;
-
-          end_range = strtoll (p, &endptr, 10);
-
-          if (end_range < start_range)
-            goto invalid_input;
-
-          p = endptr;
-        }
-
-      /* Just set some limit.  */
-      if (end_range > (1 << 20))
-        goto invalid_input;
-
-      if (end_range >= (long long) (mask_size * CHAR_BIT))
-        {
-          size_t new_mask_size = (end_range / CHAR_BIT) + 1;
-          mask = xrealloc (mask, new_mask_size);
-          memset (mask + mask_size, 0, new_mask_size - mask_size);
-          mask_size = new_mask_size;
-        }
-
-      for (long long i = start_range; i <= end_range; i++)
-        mask[i / CHAR_BIT] |= (1 << (i % CHAR_BIT));
-
-      if (*p == ',')
-        p++;
-      else if (*p)
-        goto invalid_input;
-    }
-
-  *out = mask;
-  mask = NULL;
-  *out_size = mask_size;
-
-  return 0;
-
-invalid_input:
-  return crun_make_error (err, 0, "cannot parse input `%s`", str);
-}
-
 static void
 get_systemd_scope_and_slice (const char *id, const char *cgroup_path, char **scope, char **slice)
 {
@@ -1334,7 +1266,17 @@ append_devices (sd_bus_message *m,
   size_t i;
 
   if (has_allow_all (resources->devices, resources->devices_len))
-    return 0;
+    {
+      sd_err = sd_bus_message_append (m, "(sv)", "DevicePolicy", "s", "auto");
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append DevicePolicy");
+
+      sd_err = sd_bus_message_append (m, "(sv)", "DeviceAllow", "a(ss)", 0);
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append empty DeviceAllow");
+
+      return 0;
+    }
 
   sd_err = sd_bus_message_append (m, "(sv)", "DevicePolicy", "s", "strict");
   if (UNLIKELY (sd_err < 0))
@@ -1400,6 +1342,7 @@ append_devices (sd_bus_message *m,
 
 static int
 append_resources (sd_bus_message *m,
+                  bool is_update,
                   const char *state_dir,
                   runtime_spec_schema_config_linux_resources *resources,
                   int cgroup_mode,
@@ -1544,7 +1487,13 @@ append_resources (sd_bus_message *m,
 #  undef APPEND_UINT64
 #  undef APPEND_UINT64_VALUE
 
-  return append_devices (m, resources, err);
+  if (! is_update || resources->devices)
+    {
+      ret = append_devices (m, resources, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+  return 0;
 }
 
 static int
@@ -1716,7 +1665,7 @@ enter_systemd_cgroup_scope (runtime_spec_schema_config_linux_resources *resource
         }
     }
 
-  ret = append_resources (m, state_dir, resources, cgroup_mode, err);
+  ret = append_resources (m, false, state_dir, resources, cgroup_mode, err);
   if (UNLIKELY (ret < 0))
     goto exit;
 
@@ -2026,7 +1975,7 @@ libcrun_update_resources_systemd (struct libcrun_cgroup_status *cgroup_status,
       goto exit;
     }
 
-  ret = append_resources (m, state_dir, resources, cgroup_mode, err);
+  ret = append_resources (m, true, state_dir, resources, cgroup_mode, err);
   if (UNLIKELY (ret < 0))
     goto exit;
 
